import java.awt.*;
import java.util.*;

class String2D extends ArrayList<Point> {
    public String2D(Point p) {
        firstPoint = (Point)p.clone();
	lastPoint = (Point)p.clone();
    }

    public boolean add(Point o) {
        // only add vector when its length hits four, i.e.: delta = 4
	if (Math.max(Math.abs(o.x-lastPoint.x), Math.abs(o.y-lastPoint.y)) > 4) {
	    boolean b = super.add(new Point(o.x-lastPoint.x, o.y-lastPoint.y));
	    lastPoint.setLocation(o.x, o.y);
	    return b;
	} else {
	    return false;
	}
    }
    
    public void addLine(Point p) {
        addLine(lastPoint.x, lastPoint.y, p.x, p.y);
    }

    // standard Bresenham line drawing algorithm.
    // will append the vectorial representation of the line directly 
    // without storing the point positions. only the lastPoint is kept
    // for next call...
    private void addLine(int x0, int y0, int x1, int y1) {
//    System.out.println("drawing (" + x0 + "," + y0 + ") -> (" + x1 + "," + y1 + ")");
        assert x0 != x1 || y0 != y1;

        if (lastPoint == null)
	    lastPoint = new Point(x0, y0);
	else
	    assert lastPoint.x == x0 && lastPoint.y == y0;

        int dy = y1 - y0;
        int dx = x1 - x0;
        int stepx, stepy;

        if (dy < 0) { dy = -dy;  stepy = -1; } else { stepy = 1; }
        if (dx < 0) { dx = -dx;  stepx = -1; } else { stepx = 1; }
        dy <<= 1;                                                  // dy is now 2*dy
        dx <<= 1;                                                  // dx is now 2*dx

//        add(new Point(x0, y0));
        if (dx > dy) {
            int fraction = dy - (dx >> 1);                         // same as 2*dy - dx
            while (x0 != x1) {
                if (fraction >= 0) {
                    y0 += stepy;
                    fraction -= dx;                                // same as fraction -= 2*dx
                }
                x0 += stepx;
                fraction += dy;                                    // same as fraction -= 2*dy
                add(new Point(x0, y0));
            }
        } else {
            int fraction = dx - (dy >> 1);
            while (y0 != y1) {
                if (fraction >= 0) {
                    x0 += stepx;
                    fraction -= dy;
                }
                y0 += stepy;
                fraction += dx;
                add(new Point(x0, y0));
            }
        }
    }

    private Point firstPoint, lastPoint;

    // dynamic programming method used for finding distance between strings.
    public Sequence2D distance(String2D y) {
        String2D x = this;
        int   [][] D     = new    int[x.size()+1][y.size()+1];
	Edit2D[][] edits = new Edit2D[x.size()+1][y.size()+1];

	// compute the distance table and the edit table
	for (int i = 0; i < x.size()+1; i++) {
	    for (int j = 0; j < y.size()+1; j++) {
	        if (i == 0 && j == 0) {
	            D    [i][j] = 0;
		    edits[i][j] = null;
		} else if (i == 0) {
		    Edit2D e    = new Edit2D(null, y.get(j-1));
		    D    [i][j] = D[0][j-1] + e.cost();
		    edits[i][j] = e;
		} else if (j == 0) {
		    Edit2D e    = new Edit2D(x.get(i-1), null);
		    D    [i][j] = D[i-1][0] + e.cost();
		    edits[i][j] = e;
		} else {
		    Edit2D e1 = new Edit2D(x.get(i-1), y.get(j-1));
		    Edit2D e2 = new Edit2D(x.get(i-1), null);
		    Edit2D e3 = new Edit2D(null,       y.get(j-1));

		    int d1 = D[i-1][j-1] + e1.cost();
		    int d2 = D[i-1][j  ] + e2.cost();
		    int d3 = D[i  ][j-1] + e3.cost();

		    // prefer substitutions over deletions and insertions
		    if (d1 <= d2 && d1 <= d3) {
		        D    [i][j] = d1;
			edits[i][j] = e1;
		    } else if (d2 < d1 && d2 < d3) {
		        D    [i][j] = d2;
			edits[i][j] = e2;
		    } else {
		        assert (d3 < d1 && d3 <= d2) || (d3 <= d1 && d3 < d2);
		        D    [i][j] = d3;
			edits[i][j] = e3;
		    }
		}
	    }
   	}

	// using the distance table, find path, and using path, 
	// find the collection of edits that make up the
	// sequence we are interested in
	int i = x.size();
	int j = y.size();
	Sequence2D S = new Sequence2D();
	while (true) {
	    if (i == 0 && j == 0) {
	        break; // no more edits
            } else {
	        // add edit to our list
	        S.add(edits[i][j]);
	    }

	    if (i == 0) {
	        j--; // go back 1 in y 
	    } else if (j == 0) {
	        i--; // go back 1 in x 
	    } else {
	        Edit2D e = edits[i][j];
		if (e.from == null) { // insertion
		    j--; // go back 1 in y 
		} else if (e.to == null) { // deletion
		    i--; // go back 1 in x 
		} else { // substitution
		    i--; // go back 1 in x and y
		    j--;
		}
	    }
	}
	// we accumulated the edits in reversed order
	Collections.reverse(S);
	return S;
    }

    // apply the sequence of edits to this string
    public void apply(Sequence2D S) {
	ListIterator<Point>   it =   listIterator(0);
	ListIterator<Edit2D> Sit = S.listIterator();
	while (Sit.hasNext()) {
	    Edit2D e = Sit.next();
	    if (e.from == null) {
	        it.add(e.to); // insertion
		              // automatically goes to next symbol
	    } else if (e.to == null) {
	        Point p = it.next();
	        assert p.equals(e.from);
	        it.remove();  // deletion
		              // iterator will point to next symbol
	    } else {
	        Point p = it.next();
	        assert p.equals(e.from);
	        it.set(e.to); // substitution
		              // automatically goes to next symbol
	    }
	}
    }

    // draw this string segment by segment
    void paint(Graphics g) {
	int x0 = firstPoint.x,
	    y0 = firstPoint.y;
	Iterator<Point> it = iterator();
	while (it.hasNext()) {
	    Point p = it.next();
	    int x1 = x0 + p.x;
	    int y1 = y0 + p.y;
	    g.drawLine(x0, y0, x1, y1);
	    x0 = x1;
	    y0 = y1;
	}
    }
    
    public Object clone() {
        String2D c = new String2D(firstPoint);
	c.addAll(this);
	c.firstPoint = (Point)firstPoint.clone();
	c.lastPoint = (Point)lastPoint.clone();
	return c;
    }
}
